<?php

/**
 * @file
 * Contains Drupal\scheduled_updates\Controller\ScheduledUpdateAddController.
 */

namespace Drupal\scheduled_updates\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Url;
use Drupal\scheduled_updates\ScheduledUpdateTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;


/**
 * Class ScheduledUpdateAddController.
 *
 * @package Drupal\scheduled_updates\Controller
 */
class ScheduledUpdateAddController extends ControllerBase {

  /** @var \Drupal\Core\Entity\EntityStorageInterface  */
  protected $typeStorage;

  /**
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $storage;

    public function __construct(EntityStorageInterface $storage, EntityStorageInterface $type_storage) {
      $this->storage = $storage;
      $this->typeStorage = $type_storage;
    }

    /**
     * {@inheritdoc}
     */
    public static function create(ContainerInterface $container) {
      /** @var EntityManagerInterface $entity_manager */
      $entity_manager = $container->get('entity.manager');
      return new static(
        $entity_manager->getStorage('scheduled_update'),
        $entity_manager->getStorage('scheduled_update_type')
      );
    }
    /**
     * Displays add links for available bundles/types for entity
     * scheduled_update .
     *
     * @param \Symfony\Component\HttpFoundation\Request $request
     *   The current request object.
     *
     * @return array
     *   A render array for a list of the scheduled_update bundles/types that
     *   can be added or if there is only one type/bunlde defined for the site,
     *   the function returns the add page for that bundle/type.
     */
    public function add(Request $request) {
      $types = $this->getReferencingTypes();
      if ($types && count($types) == 1) {
        $type = reset($types);
        return $this->addForm($type, $request);
      }
      if (count($types) === 0) {
        return array(
          '#markup' => $this->t('You have not created any %bundle types yet. @link to add a new type.', [
            '%bundle' => 'Scheduled update',
            '@link' => $this->l($this->t('Go to the type creation page'), Url::fromRoute('entity.scheduled_update_type.add_form')),
          ]),
        );
      }
      return array('#theme' => 'scheduled_update_content_add_list', '#content' => $types);
    }

    /**
     * Presents the creation form for scheduled_update entities of given
     * bundle/type.
     *
     * @param EntityInterface $scheduled_update_type
     *   The custom bundle to add.
     * @param \Symfony\Component\HttpFoundation\Request $request
     *   The current request object.
     *
     * @return array
     *   A form array as expected by drupal_render().
     */
    public function addForm(EntityInterface $scheduled_update_type, Request $request) {
      $entity = $this->storage->create(array(
        'type' => $scheduled_update_type->id()
      ));
      return $this->entityFormBuilder()->getForm($entity);
    }

    /**
     * Provides the page title for this controller.
     *
     * @param EntityInterface $scheduled_update_type
     *   The custom bundle/type being added.
     *
     * @return string
     *   The page title.
     */
    public function getAddFormTitle(EntityInterface $scheduled_update_type) {
      return t('Create of bundle @label',
        array('@label' => $scheduled_update_type->label())
      );
    }

  protected function getReferencingTypes() {
    $types = [];
    /** @var ScheduledUpdateTypeInterface $type */
    foreach ($this->typeStorage->loadMultiple() as $type) {
      if ($type->isReferencerType()) {
        $types[] = $type;
      }
    }
    return $types;
  }

  /**
   * Determine access to update add page.
   *
   * If user has permission to add any types they should have access to this page.
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *
   * @return \Drupal\Core\Access\AccessResult
   */
  public function addPageAccess(AccountInterface $account) {
    $types = $this->typeStorage->loadMultiple();
    $perms = [];
    foreach ($types as $type_id => $type) {
      $perms[] = "create $type_id scheduled updates";
    }
    return AccessResult::allowedIfHasPermissions($account, $perms, 'OR');
  }

}
