<?php
/**
 * @file
 * Contains \Drupal\scheduled_updates\Plugin\BaseUpdateRunner.
 */


namespace Drupal\scheduled_updates\Plugin;


use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\scheduled_updates\ClassUtilsTrait;
use Drupal\scheduled_updates\Entity\ScheduledUpdate;
use Drupal\scheduled_updates\Entity\ScheduledUpdateType;
use Drupal\scheduled_updates\ScheduledUpdateInterface;
use Drupal\scheduled_updates\ScheduledUpdateTypeInterface;
use Drupal\scheduled_updates\RevisionUtils;

use Symfony\Component\DependencyInjection\ContainerInterface;

abstract class BaseUpdateRunner extends PluginBase implements UpdateRunnerInterface{

  use ClassUtilsTrait;

  /** @var \Drupal\Core\Entity\EntityTypeManagerInterface  */
  protected $entityTypeManager;
  /** @var \Drupal\Core\Entity\EntityFieldManagerInterface  */
  protected $fieldManager;

  /**
   * @var \Drupal\scheduled_updates\RevisionUtils
   */
  protected $revisionUtils;

  /**
   * @var \Drupal\Core\Session\AccountSwitcherInterface
   */
  protected $accountSwitcher;

  /** @var  \Drupal\scheduled_updates\entity\ScheduledUpdateType */
  protected $scheduled_update_type;


  /**
   * Queue items that will be released after updates in queue are run.
   *
   * The items were not valid updates.
   *
   * @var array
   */
  protected $items_to_release;

  /**
   * If the runner is currently switched to a different user.
   * @var boolean
   */
  protected $isUserSwitched;


  /**
   * @return mixed
   */
  public function getAfterRun() {
    return $this->configuration['after_run'];
  }

  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityFieldManagerInterface $fieldManager, EntityTypeManagerInterface $entityTypeManager, RevisionUtils $revisionUtils, AccountSwitcherInterface $accountSwitcher) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->fieldManager = $fieldManager;
    $this->entityTypeManager = $entityTypeManager;
    $this->scheduled_update_type = ScheduledUpdateType::load($configuration['updater_type']);
    $this->revisionUtils = $revisionUtils;
    $this->accountSwitcher = $accountSwitcher;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager'),
      $container->get('scheduled_updates.type_info'),
      $container->get('account_switcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function addUpdatesToQueue() {
    $updates = $this->getAllUpdates();
    if ($updates) {
      $queue = $this->getQueue();

      foreach ($updates as $update) {
        $queue->createItem($update);
        /** @var ScheduledUpdate $update */
        $update = ScheduledUpdate::load($update['update_id']);
        $update->status = ScheduledUpdateInterface::STATUS_INQUEUE;
        $update->save();
      }
    }
  }

  /**
   * Get all schedule updates.
   *
   * @return ScheduledUpdate[]
   */
  protected function getAllUpdates() {
    $updates = [];
    /** @var ScheduledUpdate[] $updates */
    switch ($this->scheduled_update_type->getReferenceType()) {
      case ScheduledUpdateTypeInterface::REFERENCED:
        $updates = $this->getReferencedUpdates();
        break;
      case ScheduledUpdateTypeInterface::REFERENCER:
        $updates = $this->getReferencingUpdates();
        break;
    }
    return $updates;
  }

  /**
   * Return all schedule updates that are referenced via Entity Reference
   * fields.
   *
   * @return ScheduledUpdate[]
   */
  protected function getReferencedUpdates() {
    $updates = [];
    /** @var String[] $fields */
    if ($field_ids = $this->getReferencingFieldIds()) {

      //$query = $entity_storage->getQuery('OR');
      $entity_ids = [];

      $entity_storage = $this->entityTypeManager->getStorage($this->updateEntityType());
      foreach ($field_ids as $field_id) {

        $query = $entity_storage->getQuery('AND');
        $this->addActiveUpdateConditions($query, "$field_id.entity.");
        $entity_ids += $query->execute();
      }
      if ($entities = $entity_storage->loadMultiple($entity_ids)) {
        /** @var ContentEntityInterface $entity */
        foreach ($entities as $entity) {
          /** @var  $entity_update_ids - all update ids for this entity for our fields. */
          $entity_update_ids = [];
          /** @var  $field_update_ids - update ids keyed by field_id. */
          $field_update_ids = [];
          foreach ($field_ids as $field_id) {
            // Store with field id.
            $field_update_ids[$field_id] = $this->getEntityReferenceTargetIds($entity, $field_id);
            // Add to all for entity.
            $entity_update_ids += $field_update_ids[$field_id];
          }
          // For all entity updates return only those ready to run.
          $ready_update_ids = $this->getReadyUpdates($entity_update_ids);
          // Loop through updates attached to fields.
          foreach ($field_update_ids as $field_id => $update_ids) {
            // For updates attached to field get only those ready to run.
            $field_ready_update_ids = array_intersect($update_ids, $ready_update_ids);
            foreach ($field_ready_update_ids as $field_ready_update_id) {
              $updates[] = [
                'update_id' => $field_ready_update_id,
                'entity_ids' => [$entity->id()],
                'field_id' => $field_id,
                'entity_type' => $this->updateEntityType(),
              ];
            }
          }
        }
      }
    }
    return $updates;

  }

  /**
   * Get all scheduled updates that referencing entities via Entity Reference
   * Field
   *
   *  @return ScheduledUpdate[]
   */
  protected function getReferencingUpdates() {
    $updates = [];
    $update_ids = $this->getReadyUpdates();
    foreach ($update_ids as $update_id) {
      $updates[] = [
        'update_id' => $update_id,
        'entity_type' => $this->updateEntityType(),
      ];
    }
    return $updates;
  }

  /**
   * {@inheritdoc}
   */
  public function runUpdatesInQueue($items_to_run = 30) {
    $queue = $this->getQueue();

    for ($item_num = 0; $item_num < $items_to_run; $item_num++) {
      if ($update_info = $queue->claimItem(600)) {
        /** @var ScheduledUpdate $update */
        if ($update = ScheduledUpdate::load($update_info->data['update_id'])) {
          if (isset($update_info->data['entity_ids'])) {
            $entity_ids = $update_info->data['entity_ids'];
          }
          else {
            $entity_ids = $update->getUpdateEntityIds();
          }
          $this->runUpdate($update, $entity_ids, $update_info);

        }
        else {
          // Could not load update that is in queue.
          $queue->deleteItem($update_info);
        }
      }
      else {
        // No more items can be claimed.
        break;
      }

    }
    $this->releaseClaimedItems();

  }

  /**
   * Run an individual update from the queue.
   *
   * The update may involve multiple entities.
   *
   * @param \Drupal\scheduled_updates\Entity\ScheduledUpdate $update
   *
   * @param array $entity_ids
   *  Ids of entities that should be updated
   * @param $queue_item
   *
   * @return bool
   */
  protected function runUpdate(ScheduledUpdate $update, $entity_ids, $queue_item) {
    $entity_storage = $this->entityTypeManager->getStorage($this->updateEntityType());

    /** @var ContentEntityInterface[] $entities_to_update */
    $entities_to_update = $entity_storage->loadMultiple($entity_ids);

    $update_count = 0;
    $invalid_count = 0;
    $invalid_entity_ids = [];
    foreach (array_keys($entities_to_update) as $entity_id) {
      $entity_to_update = $entities_to_update[$entity_id];
      $this->prepareEntityForUpdate($update, $queue_item, $entity_to_update);
      $this->switchUser($update, $entity_to_update);
      $violations = $entity_to_update->validate();
      if (count($violations)) {
        $invalid_entity_ids[] = $entity_id;
      }
      else {
        // Validation was successful.
        $entity_to_update->save();
        $update_count++;
      }
      $this->switchUserBack();
    }

    $this->display_message($update_count, $invalid_count);
    $successful = empty($invalid_entity_ids);
    if (!$successful) {
      // At least 1 entity could not be updated.
      if ($this->getInvalidUpdateBehavior() == UpdateRunnerInterface::INVALID_REQUEUE) {
        // We can't release the item now or it will be claimed again.
        $update->status = ScheduledUpdateInterface::STATUS_REQUEUED;
        if ($this->scheduled_update_type->isReferencerType()) {
          $update->setUpdateEntityIds($invalid_entity_ids);
        }
        $update->save();
        $this->addItemToRelease($queue_item);
      }
      elseif ($this->getInvalidUpdateBehavior() == UpdateRunnerInterface::INVALID_ARCHIVE) {
        // @todo Should the successful entities be removed
        $update->status = ScheduledUpdateInterface::STATUS_UNSUCESSFUL;
        $update->save();
        $this->getQueue()->deleteItem($queue_item);
      }
      else {
        $update->delete();
        $this->getQueue()->deleteItem($queue_item);
      }
    }
    else {
      // There were no invalid entity updates
      if ($this->getAfterRun() == UpdateRunnerInterface::AFTER_DELETE) {
        $update->delete();
      }
      else {
        $update->status = ScheduledUpdateInterface::STATUS_SUCCESSFUL;
        $update->save();
      }
      $this->getQueue()->deleteItem($queue_item);
    }
    return $successful;
  }

  /**
   * @return array
   */
  protected function getReferencingFieldIds() {
    $field_ids = [];
    $entity_reference_fields = $this->fieldManager->getFieldMapByFieldType('entity_reference');
    $update_entity_type = $this->updateEntityType();
    if (empty($entity_reference_fields[$update_entity_type])) {
      return $field_ids;
    }
    $entity_reference_fields = $entity_reference_fields[$update_entity_type];
    foreach ($entity_reference_fields as $field_id => $entity_reference_field) {
      foreach ($entity_reference_field['bundles'] as $bundle) {

        $field = $this->fieldManager->getFieldDefinitions($update_entity_type, $bundle)[$field_id];
        if ($field instanceof FieldConfig) {

          if ($field->getSetting('target_type') == 'scheduled_update'
            && in_array($this->configuration['updater_type'], $field->getSetting('handler_settings')['target_bundles'])
          ) {
            $field_ids[] = $field_id;
          }
        }

      }
    }
    return $field_ids;
  }

  /**
   * @return string
   */
  protected function updateEntityType() {
    return $this->scheduled_update_type->getUpdateEntityType();
  }

  /**
   * @param $update
   * @param $entity_to_update
   */
  protected function transferFieldValues(ScheduledUpdate $update, ContentEntityInterface $entity_to_update) {
    $field_map = $this->scheduled_update_type->getFieldMap();
    foreach ($field_map as $from_field => $to_field) {
      if ($to_field) {
        if ($entity_to_update->hasField($to_field) && $update->hasField($from_field)) {
          $entity_to_update->set($to_field, $update->get($from_field)->getValue());
        }
      }
    }
  }

  protected function display_message($update_count, $invalid_count) {
    $msg = $this->t('Updater <em>@title</em> complete. Results:', ['@title' => $this->scheduled_update_type->label()]);
    if ($update_count) {
      $msg .= '<br/>' . $this->t('@count update(s) were performed.', ['@count' => $update_count]);
    }
    else {
      $msg .= '<br/>' . $this->t('No updates were performed.');
    }
    if ($invalid_count) {
      $msg .= '<br/>' . $this->t('@count updates were invalid', ['@count' => $invalid_count]);
    }
    drupal_set_message($msg);
  }

  /**
   * {@inheritdoc}
   */
  public function getQueue() {
    return \Drupal::queue('scheduled_updates:' . $this->configuration['updater_type'],TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $settings = $form_state->get('update_runner');

    /** @var ScheduledUpdateType $type */
    $type = $form_state->get('scheduled_update_type');

    $form['id'] = [
      '#type' => 'value',
      '#value' => $this->getPluginId(),
    ];
    $form['after_run'] = [
      '#type' => 'select',
      '#title' => $this->t('After update behavior'),
      '#description' => $this->t('What should happen after updates are run?'),
      '#required' => TRUE,
      '#default_value' => $settings['after_run']? $settings['after_run']: UpdateRunnerInterface::AFTER_DELETE,
      '#options' => [
        UpdateRunnerInterface::AFTER_DELETE => $this->t('Delete Updates'),
        UpdateRunnerInterface::AFTER_ARCHIVE => $this->t('Archive Updates'),
      ],
    ];
    $form['invalid_update_behavior'] = [
      '#type' => 'select',
      '#title' => $this->t('Invalid update behavior'),
      '#description' => $this->t('What should happen when an update is invalid?'),
      '#required' => TRUE,
      '#default_value' => $settings['invalid_update_behavior']? $settings['invalid_update_behavior']: UpdateRunnerInterface::INVALID_DELETE,
      '#options' => [
        UpdateRunnerInterface::INVALID_REQUEUE => $this->t('Leave update in queue'),
        UpdateRunnerInterface::INVALID_DELETE => $this->t('Delete update'),
        UpdateRunnerInterface::INVALID_ARCHIVE => $this->t('Archive update'),
       ],
    ];

    if ($type && $this->revisionUtils->supportsRevisionUpdates($type)) {
      $revision_options = [];
      if ($this->revisionUtils->supportsRevisionBundleDefault($type)) {
        $revision_options = [
          UpdateRunnerInterface::REVISIONS_BUNDLE_DEFAULT => $this->t('Use bundle default.'),
        ];
      }
      $revision_options += [
        UpdateRunnerInterface::REVISIONS_YES => $this->t('Always Create New Revisions'),
        UpdateRunnerInterface::REVISIONS_NO => $this->t('Never Create New Revisions'),
      ];
      $form['create_revisions'] = [
        '#type' => 'select',
        '#title' => $this->t('Create New Revisions'),
        '#description' => $this->t('Should updates create new revisions of entities? Not all entity types support revisions.'),
        '#required' => TRUE,
        '#default_value' => $settings['create_revisions']? $settings['create_revisions']: UpdateRunnerInterface::REVISIONS_BUNDLE_DEFAULT,
        '#options' => $revision_options,
      ];
    }
    $update_user_options = [
      UpdateRunnerInterface::USER_UPDATE_RUNNER => $this->t('The user who using running the updates. User #1 in cron.'),
      UpdateRunnerInterface::USER_UPDATE_OWNER => $this->t('The owner of the update.'),
    ];

    if ($type && $this->revisionUtils->supportsOwner($type)) {
      $update_user_options[UpdateRunnerInterface::USER_OWNER] = $this->t('The owner of the entity to be updated');
    }
    if ($type && $this->revisionUtils->supportsRevisionOwner($type)) {
      $update_user_options[UpdateRunnerInterface::USER_REVISION_OWNER] = $this->t('The owner of the last revision.');
    }
    $form['update_user'] = [
      '#type' => 'select',
      '#title' => $this->t('Run update as:'),
      '#description' => $this->t('Which user should the updates be run as?'),
      '#required' => TRUE,
      '#default_value' => $settings['update_user']? $settings['update_user']: UpdateRunnerInterface::USER_UPDATE_RUNNER,
      '#options' => $update_user_options,
    ];



    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getInvalidUpdateBehavior() {
    return $this->configuration['invalid_update_behavior'];
  }


  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
  }

  /**
   * @param $queue_item
   */
  protected function addItemToRelease($queue_item) {
    $this->items_to_release[] = $queue_item;
  }

  protected function releaseClaimedItems() {
    $queue = $this->getQueue();
    foreach ($this->items_to_release as $item) {
      $queue->releaseItem($item);
    }
  }

  /**
   * Add conditions to a query to select updates to run.
   *
   * @param \Drupal\Core\Entity\Query\QueryInterface $query
   * @param string $condition_prefix
   *  String to attach to the beginning of all conditions if the base table is not updates.
   */
  protected function addActiveUpdateConditions(QueryInterface $query, $condition_prefix = '') {
    $query->condition($condition_prefix. 'update_timestamp', REQUEST_TIME, '<=');
    $query->condition($condition_prefix. 'type', $this->scheduled_update_type->id());
    $query->condition(
      $condition_prefix. 'status',
      [
        ScheduledUpdateInterface::STATUS_UNRUN,
        ScheduledUpdateInterface::STATUS_REQUEUED,
      ],
      'IN'
    );
  }

  /**
   * Remove update from reference field value.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   * @param \Drupal\scheduled_updates\Entity\ScheduledUpdate $update
   * @param $field_id
   */
  protected function removeUpdate(ContentEntityInterface $entity, ScheduledUpdate $update, $field_id) {
    $current_update_ids = $this->getEntityReferenceTargetIds($entity,$field_id);
    $new_update_ids = array_diff($current_update_ids, [$update->id()]);
    $entity->get($field_id)->setValue($new_update_ids);
  }

  /**
   * @todo Is there a way to do this with core Field API?
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   * @param $field_name
   *
   * @return array
   */
  protected function getEntityReferenceTargetIds(ContentEntityInterface $entity, $field_name) {
    $field_values = $entity->get($field_name)->getValue();
    $target_ids = [];
    foreach ($field_values as $field_value) {
      $target_ids[] = $field_value['target_id'];
    }
    return $target_ids;
  }

  /**
   * Get updates that are ready to be run for this Runner.
   *
   * @param array $update_ids
   *  If given updates will be restrict to this array.
   *
   * @return array
   *  Update ids.
   */
  protected  function getReadyUpdates($update_ids = []) {
    $entity_storage = $this->entityTypeManager->getStorage('scheduled_update');
    $query = $entity_storage->getQuery('AND');
    $this->addActiveUpdateConditions($query);
    if ($update_ids) {
      $query->condition('id', $update_ids, 'IN');
    }
    return $query->execute();
  }

  /**
   * Prepare an entity to be updated.
   *
   * @param \Drupal\scheduled_updates\Entity\ScheduledUpdate $update
   * @param $queue_item
   * @param ContentEntityInterface $entity_to_update
   */
  protected function prepareEntityForUpdate(ScheduledUpdate $update, $queue_item, ContentEntityInterface $entity_to_update) {
    $this->transferFieldValues($update, $entity_to_update);
    if (!empty($queue_item->data['field_id'])) {
      $this->removeUpdate($entity_to_update, $update, $queue_item->data['field_id']);
    }
    $this->setEntityRevision($entity_to_update, $update);
  }

  /**
   * Set a entity to use a new revision is applicable.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity_to_update
   * @param \Drupal\scheduled_updates\Entity\ScheduledUpdate $update
   */
  protected function setEntityRevision(ContentEntityInterface $entity_to_update, ScheduledUpdate $update) {
    if ($this->revisionUtils->isRevisionableUpdate($update)) {
      $new_revision = FALSE;
      $create_revisions = $this->configuration['create_revisions'];
      if ($create_revisions == UpdateRunnerInterface::REVISIONS_BUNDLE_DEFAULT) {
        $new_revision = $this->revisionUtils->getRevisionDefault($entity_to_update);
      }
      elseif ($create_revisions == UpdateRunnerInterface::REVISIONS_YES) {
        $new_revision = TRUE;
      }
      $entity_to_update->setNewRevision($new_revision);
      if ($new_revision) {
        $this->revisionUtils->setRevisionCreationTime($entity_to_update);
      }
    }
  }

  /**
   * Switch to another user to run an update if necessary.
   *
   * @param \Drupal\scheduled_updates\Entity\ScheduledUpdate $update
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity_to_update
   */
  protected function switchUser(ScheduledUpdate $update, ContentEntityInterface $entity_to_update) {
    $update_user = $this->configuration['update_user'];
    $switch_to_user = NULL;
    switch ($update_user) {
      case $this::USER_UPDATE_RUNNER:
        // Running the update as the update runner means there is no need to switch
        return;
      case $this::USER_OWNER:
        $switch_to_user = $this->getEntityOwner($entity_to_update);
        break;
      case $this::USER_REVISION_OWNER:
        $switch_to_user = $this->getRevisionOwner($entity_to_update);
        break;
      case $this::USER_UPDATE_OWNER:
        $switch_to_user = $update->getOwner();
        break;
    }
    if ($switch_to_user) {
      // @todo Throw an error because we should have a user.
      $this->accountSwitcher->switchTo($switch_to_user);
      $this->isUserSwitched = TRUE;
    }

  }

  protected function switchUserBack() {
    if ($this->isUserSwitched) {
      $this->accountSwitcher->switchBack();
      $this->isUserSwitched = FALSE;
    }
  }

}
